//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File  main.cpp
// Date 26.01.08 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//--------------------------------------------------------------------------
// Programm part for Arduino-Diecimila
//***************************************************************************

//***************************************************************************
// Notice
//***************************************************************************

/*
    int's on avr platform are 2 bytes long
    longs's are 4 bytes long
*/

//***************************************************************************
// Includes
//***************************************************************************

#include <WProgram.h>

//***************************************************************************
// Settings
//***************************************************************************

#define SPI_EXTENSION               // comment out if no SPI extension used 
                                    // (without external shift registers for more io)

#define GHOSTCAR                    // comment out if no GHOST-CAR used 

#define TIMER_CLOCK_FREQ 2000000.0  // 2MHz for /8 prescale from 16MHz

//***************************************************************************
// Input Value
//***************************************************************************

struct IoValue
{
   unsigned long time;        // time
   unsigned long value;       // value
};

//***************************************************************************
// class FifoCache
//
//  - Small 'Interrupt Safe' FIFO Cache
//***************************************************************************

class FifoCache
{
   public:

      // declarations

      struct Node
      {
         Node* prev;
         IoValue data;
         Node* next;
      };

      FifoCache()   { first = _first = last = _last = 0; cnt = _cnt = 0; locked = false; }
      ~FifoCache()  { flush(); }

      // interface

      byte count()            { return cnt; }
      byte full()             { return cnt >= 100; }
      void lock()             { locked = true; }
      void unlock()           { locked = false; }

      // push called from inside the interrupt

      void push(unsigned long value, unsigned long time) 
      {
         if (!locked)
         {
            // first move internal list to public

            while (_cnt)
            {
               IoValue v = _pop(_first, _last, _cnt);
               _push(v.value, v.time, first, last, cnt);
            }

            _push(value, time, first, last, cnt);
         }
         else
         {
            _push(value, time, _first, _last, _cnt);
         }
      }

      // pop and flush called form extern

      IoValue pop()
      { 
         IoValue v;

         lock();
         v = _pop(first, last, cnt);
         unlock();

         return v;
      }

      void flush()    
      { 
         lock();
         while (cnt) pop();
         unlock();
      }

   private:

      // functions

      void _push(unsigned long value, unsigned long time, Node*& f, Node*& l, byte& c)
      {
         Node* n = (Node*)malloc(sizeof(Node));

         n->data.value = value;
         n->data.time = time;

         n->prev = l;
         n->next = 0;

         if (l)  l->next = n;
         l = n;
         if (!f) f = l;

         c++;
      }

      IoValue _pop(Node*& f, Node*& l, byte& c)
      { 
         IoValue v; 

         if (!f)
         {
            c = 0;
            v.time = 0;
            v.value = 0;
            return v;
         }

         v = f->data;

         Node* tmp = f;

         if (f->next)
         {
            f = f->next;
            f->prev = 0;
         }

         free(tmp);
         c--;

         return v; 
      }

      // data

      volatile byte locked;

      byte cnt;
      Node* first;
      Node* last;

      byte _cnt;
      Node* _first;
      Node* _last;
};

//**************************************************************
// Definitions
//**************************************************************

enum Misc
{
   fail        = -1,
   success     = 0,
   na          = -1,
   TB          = 1,

   maxCmd      = 100,
   cmdEnd      = ';',
   bounceTime  = 200,

   firstInBit         = 2,
   lastInBit          = 4,

#ifdef GHOSTCAR

   pmwOut1            = 5,
   pmwOut2            = 6,

#endif

   firstOutBit        = 7,
   lastOutBit         = 9,

#ifdef SPI_EXTENSION

   firstExtendedOut   = 16,

   // the SPI bus

   bitShiftRegSlave   = 10,
   bitShiftRegDataOut = 11,
   bitShiftRegDataIn  = 12,
   bitShiftRegClock   = 13,

#endif
};

//**************************************************************
// Globals
//**************************************************************

FifoCache inputCache;                     // cache for digital input
volatile unsigned int outputValue = 0;    // value for output via shift-register

volatile int timerLoadValue = 0;
volatile int ghostcarPin = na;

//***************************************************************************
// Prototypes
//***************************************************************************

void sendAnalogValue(byte pin);
unsigned char setupTimer2(float irqFreq);
void setupSpiBus();

//***************************************************************************
// Debug
//***************************************************************************

void debug(const char* msg)
{
  Serial.print("DEBUG ");
  Serial.print(msg);
  Serial.println(";");
}

void debug(unsigned long value, const char* msg = 0)
{
  Serial.print("DEBUG:");

  if (msg)
     Serial.print(msg);
  else
     Serial.print("value");

  Serial.print("=0x");
  Serial.print(value, HEX);
  Serial.println(";");
}

//**************************************************************
// Setup
//**************************************************************

void setup()
{
   // initialize the serial interface

   Serial.begin(9600);

   // init digital io pins

   for (byte i = firstInBit; i <= lastInBit; i++)
   {
      pinMode(i, INPUT);
      digitalWrite(i, HIGH);  // activte the internal pullup
   }

   for (byte i = firstOutBit; i <= lastOutBit; i++)
   {
      pinMode(i, OUTPUT);
      digitalWrite(i, LOW);
   }

#ifdef GHOSTCAR
   pinMode(pmwOut1, OUTPUT);
   pinMode(pmwOut2, OUTPUT);
#endif

#ifdef SPI_EXTENSION
   setupSpiBus();
#endif

   // init interrupt stuff

   timerLoadValue = setupTimer2(10000);     // set interrupt to 10kHz

   // init some variables 

   outputValue = 0;
}

//***************************************************************************
// Setup SPI bus
//***************************************************************************

void setupSpiBus()
{
   byte clr;

   // setup SPI register control pins

   pinMode(bitShiftRegClock, OUTPUT);
   pinMode(bitShiftRegDataOut, OUTPUT);
   pinMode(bitShiftRegDataIn, INPUT);
   pinMode(bitShiftRegSlave, OUTPUT);
   digitalWrite(bitShiftRegSlave, HIGH);  // disable device

   // Now we set the SPI Control register (SPCR) to the binary value 01010000. 
   // In the control register each bit sets a different functionality. The eighth 
   // bit disables the SPI interrupt, the seventh bit enables the SPI, the sixth 
   // bit chooses transmission with the most significant bit going first, the fifth
   // bit puts the Arduino in Master mode, the fourth bit sets the data clock idle 
   // when it is low, the third bit sets the SPI to sample data on the rising edge
   // of the data clock, and the second and first bits set the speed of the SPI to 
   // system speed / 4 (the fastest). After setting our control register up we read 
   // the SPI status register (SPSR) and data register (SPDR) in to the junk clr 
   // variable to clear out any spurious data from past runs:

   SPCR = (1 << SPE) | (1 << MSTR);

   // SPI als Master, High-Bits zuerst, SCK ist HIGH wenn inaktiv 
   // SPCR = (1 << SPE) | (1 << MSTR) | (1 << CPOL);
   // maximale Geschwindigkeit: F_CPU / 2 
   // SPSR |= (1 << SPI2X);

   clr = SPSR;
   clr = SPDR;
   delay(10);
}

//***************************************************************************
// Setup Timer 2
//    Configures the 8 Bit Timer2 to generate an interrupt
//    at the specified frequency.
//    Returns the timer load value which must be loaded into TCNT2
//    inside your ISR routine.
//***************************************************************************

unsigned char setupTimer2(float irqFreq)
{
   unsigned char result; // The timer load value.

   // Calculate the timer load value
   // The 257 really should be 256 but I get better results with 257.

   result = (int)((257.0-(TIMER_CLOCK_FREQ/irqFreq))+0.5);

   // Timer2 Settings: Timer Prescaler /8, mode 0
   // Timer clock = 16MHz/8 = 2Mhz or 0.5us
   // The /8 prescale gives us a good range to work with
   // so we just hard code this for now.

   TCCR2A = 0;
   TCCR2B = 0<<CS22 | 1<<CS21 | 0<<CS20;

   TIMSK2 = 1<<TOIE2;  // Timer2 Overflow Interrupt Enable
   TCNT2 = result;     // load the timer for its first cycle

   return result;
}

//***************************************************************************
// Timer2 overflow interrupt vector handler
//***************************************************************************

ISR(TIMER2_OVF_vect) 
{
   static int keyCheckScale = 0;
   static int irqScale = 0;

   static unsigned long lastInputTimes[32];            // time for each bit used to avoid bouncing
   static unsigned long lastInputValue = 0xFFFFFFFF;   // remember last input states to avoid bouncing

   // toggle the IO pin 'firstOutBit' to test the interrupt (will be removed later)

   if (!irqScale--)
   {
      static byte value = LOW;

      // pin 'firstOutBit' flash in 1/1000 of the interrupt frequency 
      // 10kHz / 1000 = 10Hz

      value = !value;
      digitalWrite(firstOutBit, value);

      if (ghostcarPin != na)
         sendAnalogValue(ghostcarPin);

      irqScale = 1000;
   }

   // check io state

   if (!keyCheckScale--)
   {
      unsigned long mask, b;
      unsigned long inputValue = 0;
      unsigned long value = lastInputValue;
      byte changes = 0;
      byte bit;
      
      // --------------------------------
      // Support Internal IO (byte 0+1)

      for (bit = firstInBit; bit <= lastInBit; bit++)
      {
         mask = digitalRead(bit) << bit;
         inputValue |= mask;
      }

#ifdef SPI_EXTENSION

      // --------------------------------
      // Support External IO (byte 2+3)
  
      digitalWrite(bitShiftRegSlave, LOW);      // enable device

      // shift all bits in/out

      for (byte n = 0; n < 2; n++)
      {
         // start the transmission

         if (n == 0)
            SPDR = (outputValue >> 8) & 0xFF;        // heigh byte
         else
            SPDR = outputValue & 0xFF;               // low byte
         
         while (!(SPSR & (1 << SPIF)))               // wait the end of the transmission
            ;

         unsigned long v = SPDR;

         if (n == 0)
            inputValue |= ((v << 24) & 0xFF000000);  // heigh byte
         else
            inputValue |= ((v << 16) & 0x00FF0000);  // low byte
      }

      digitalWrite(bitShiftRegSlave, HIGH);          // disable device

#endif // SPI_EXTENSION

      // auswerten

      for (bit = 0; bit < 32; bit++)
      {
         mask = 1UL << bit;
         b = (inputValue & mask) ? 1 : 0;
         b = b << bit;

         if (((lastInputValue & mask) ^ b) && (millis() - lastInputTimes[bit] > bounceTime))
         {
            // bit changed -and- last change older than 'bounceTime'

            lastInputTimes[bit] = millis();
            changes++;

            if (b)
               value |= mask;
            else
               value &= ~mask;
         }
      }
      
      if (changes && !inputCache.full())
      {
         lastInputValue = value;
         inputCache.push(lastInputValue, millis());
      }

      keyCheckScale = 10;
   }

   // Capture the current timer value. This is how much error we
   // have due to interrupt latency and the work in this function

   int latency = TCNT2;

   // Reload the timer and correct for latency.

   TCNT2 = latency + timerLoadValue;
}

//**************************************************************
// Set/Clear Bit
//**************************************************************

void setBit(byte bit, volatile unsigned int& value)
{
   value |= (1 << bit);
}

void clearBit(byte bit, volatile unsigned int& value)
{
   value &= ~(1 << bit);
}

//**************************************************************
// 
//**************************************************************

void sendAnalogValue(byte pin)
{
   Serial.print("AI:");
   Serial.print(pin, DEC);
   Serial.print("=");
   Serial.print(analogRead(pin));
   Serial.println(";");
}

//**************************************************************
// 
//**************************************************************

void sendDigitalIoState()
{
   IoValue v;

   while (inputCache.count())
   {
      v = inputCache.pop();

      Serial.print("DI:");
      Serial.print(v.time);
      Serial.print("=0x");
      Serial.print(v.value, HEX);
      Serial.println(";");
   }
}

//**************************************************************
// Command GETTIME
//**************************************************************

void cmdGettime(const char* cmd)
{
  Serial.print("BOARDTIME=");
  Serial.print(millis());
  Serial.println(";");
}

//**************************************************************
// Command GETINPUTS
//**************************************************************

void cmdGetInputs(const char* cmd)
{
   sendDigitalIoState();
}

//**************************************************************
// Command DO
//**************************************************************

void cmdDigitalOut(const char* cmd)
{
   char* p = strchr(cmd, '=');
   byte pin;
   byte state;

   if (!p) return ;

   *p = 0;
   p++;

   pin = atoi(cmd);
   state = atoi(p) ? HIGH : LOW;

   if (pin >= firstExtendedOut)
   {
      pin -= firstExtendedOut;

      if (state)
         setBit(pin, outputValue);
      else
         clearBit(pin, outputValue);
   }
   else
      digitalWrite(pin, state);
}

//**************************************************************
// Command AO
//**************************************************************

void cmdAnalogOut(const char* cmd)
{
   char* p = strchr(cmd, '=');

   if (!p) return ;

   *p = 0;
   p++;

   analogWrite(atoi(cmd), atoi(p));
}

//**************************************************************
// Command RGC 'Record - Ghost Car'
//**************************************************************

void cmdRecordGhostCar(const char* cmd)
{
   unsigned int cycle;
   unsigned int pin;
   char* p;

   if (!(p = strchr(cmd, '=')))
      return ;

   p++;

   if ((cycle = atoi(p)) < 10)
      return ;
    
   if (!(p = strchr(cmd, ':')))
      return ;

   p++;

   if ((pin = atoi(p)) < 0 || pin > 5)
      return ;
    
   inputCache.flush();
   ghostcarPin = pin;

   // TODO: better we handle this in the timer interrupt

   while (!inputCache.count() && !Serial.available())
   {
      // sendAnalogValue(pin);
      // delay(cycle);
   }

   ghostcarPin = na;

   debug("RGC ready!");
}

//**************************************************************
// Command Parser
//**************************************************************

int getCommand(char* cmd)
{
   int i = 0;
   int c;
  
   while (i < maxCmd)
   {
      if ((c = Serial.read()) > 0)
      {
         if (c == cmdEnd)
            break;

         cmd[i++] = c;
      }
   }

   cmd[i] = 0;
  
   // debug(cmd);
  
   return success;
}

//**************************************************************
// Main Loop
//**************************************************************

void loop()
{
   char cmd[maxCmd+TB];
   *cmd = 0;
  
   // analogWrite(3, 128);

   // Command Dispatcher

   if (Serial.available())
   {
      getCommand(cmd);

      if (strcasecmp(cmd, "GETTIME") == 0)
         cmdGettime(cmd);

      else if (strncasecmp(cmd, "DO", 2) == 0)
         cmdDigitalOut(cmd+2);

      else if (strncasecmp(cmd, "AO", 2) == 0)
         cmdAnalogOut(cmd+2);

      else if (strncasecmp(cmd, "RGC", 3) == 0)
         cmdRecordGhostCar(cmd);

      else if (strncasecmp(cmd, "GETINPUTS", 9) == 0)
         cmdGetInputs(cmd);

   }

   // Digital IO

   if (inputCache.count())
      sendDigitalIoState();
}

//***************************************************************************
// Main
//***************************************************************************

int main()
{
	init();
	setup();

	for (;;)
		loop();

	return 0;
}
